
cdef extern from "../clib/sqrc.c":
    int isqr(int a)
    double dsqr(double a)


def sqr(a):
    """
    >>> sqr(2)
    4
    >>> '%.3g' % round(sqr(2.2), 2)
    '4.84'
    >>> sqr("asd")
    Traceback (most recent call last):
    ...
    TypeError: Expected int or float type input.
    """
    if isinstance(a, (int, )):
        return isqr(a)

    elif isinstance(a, (float, )):
        return dsqr(a)

    else:
        raise TypeError("Expected int or float type input.")



"""Discover and run doctests in Cython extension modules."""

from __future__ import annotations

import os
import pathlib
import re
from typing import Any, Iterable

import pytest
from _pytest.doctest import DoctestItem, DoctestModule
from _pytest.pathlib import ImportMode, resolve_package_path

CYTHON_SUFFIXES = [".pyx"]

IGNORE_IMPORTMISMATCH_KEY = "PY_IGNORE_IMPORTMISMATCH"
IGNORE_IMPORTMISMATCH = os.environ.get(IGNORE_IMPORTMISMATCH_KEY, "")


def pytest_addoption(parser: pytest.Parser):
    group = parser.getgroup("cython")

    group.addoption(
        "--doctest-cython",
        action="store_true",
        default=False,
        help="run doctests in all .so and .pyd modules",
        dest="doctest_cython",
    )


def pytest_collect_file(
    file_path: pathlib.Path, parent: pytest.Collector
) -> pytest.Module:
    config = parent.config
    if file_path.suffix not in CYTHON_SUFFIXES or not config.getoption(
        "--doctest-cython"
    ):
        return

    # only run test if matching .so and .pyx files exist
    return _PatchedDoctestModule.from_parent(parent, path=file_path)


class _PatchedDoctestModule(DoctestModule):
    def collect(self) -> Iterable[DoctestItem]:
        mode = ImportMode(self.config.getoption("importmode"))

        if mode is not ImportMode.importlib and IGNORE_IMPORTMISMATCH != "1":
            # we know we will get an import file mismatch error so ignore it for now,
            # we will double check the import paths after import.
            os.environ[IGNORE_IMPORTMISMATCH_KEY] = "1"

        try:
            items = list(super().collect())  # import the module and collect tests
        finally:
            # set ignore variable back to its original value
            os.environ[IGNORE_IMPORTMISMATCH_KEY] = IGNORE_IMPORTMISMATCH

        module = self.obj  # module already imported
        return _add_line_numbers(module, items)


def _without_suffixes(path: str | pathlib.Path) -> pathlib.Path:
    path = pathlib.Path(path)
    return path.with_name(path.name.split(".")[0]).with_suffix("")


def _get_module_name(path: pathlib.Path) -> str:
    pkg_path = resolve_package_path(path)
    if pkg_path is not None:
        pkg_root = pkg_path.parent
        names = list(path.with_suffix("").relative_to(pkg_root).parts)
        if names[-1] == "__init__":
            names.pop()
        module_name = ".".join(names)
    else:
        pkg_root = path.parent
        module_name = path.stem

    return module_name


def _add_line_numbers(
    module: Any, items: Iterable[DoctestItem]
) -> Iterable[DoctestItem]:
    # handle tests from Cython's internal __test__ dict generated by
    # the autotestdict directive; we exclude the tests from __test__,
    # though they do give us a little bonus if they exist: we can extract
    # the line number of the test
    lineno_re = re.compile(r"\(line (\d+)\)")
    test_dict = module.__name__ + ".__test__"
    test_items = {x.name: x for x in items}

    for test_name in list(test_items.keys()):
        if not test_name.startswith(test_dict + "."):
            continue

        match = lineno_re.search(test_name)
        lineno = int(match.group(1)) if match else None

        # If somehow the equivalent test does not already exist, we
        # keep the __test__ test (maybe it is something else not
        # generated by autotestdict)
        equiv_test_name = test_name.split()[0].replace(test_dict, module.__name__)

        if (
            equiv_test_name not in test_items
            or not test_items[equiv_test_name].dtest.examples
        ):
            # for some reason the equivalent test was not found (e.g.
            # the module was compiled with docstrings stripped) so keep
            # the __test__ test but hide the fact that it came from the
            # __test__ dict
            test_items[test_name].name = equiv_test_name
            # set lineno on the __test__ test as well, since normally
            # it is not set by doctest
            test_items[test_name].dtest.lineno = lineno
            continue

        # Delete the __test__ test, but try to update the lineno of the
        # equivalent test
        del test_items[test_name]
        test_items[equiv_test_name].dtest.lineno = lineno

    yield from sorted(test_items.values(), key=lambda x: x.name)



def foo(a, b):
    """
    >>> foo(1, 1)
    2
    """
    return a + b


class Eggs:
    def __init__(self, a, b):
        """
        >>> eggs = Eggs(1, 1)
        >>> eggs.a
        1
        >>> eggs.b
        1
        """
        self.a = a
        self.b = b

    def foo(self):
        """
        >>> eggs = Eggs(1, 1)
        >>> eggs.foo()
        2
        """
        return self.a + self.b



import pathlib

# Check imports to ensure packages are installed before running tests
import pypackage  # noqa: F401
import pytest

ROOT_PATH = pathlib.Path(__file__).parent
PROJECT_PATH = ROOT_PATH.joinpath("example-project")
PACKAGE_PATH = PROJECT_PATH.joinpath("src", "pypackage")

# TODO: Figure out if importlib can be supported.
IMPORT_MODES = ["append", "prepend"]


def get_module(basename: str, suffix=".pyx") -> pathlib.Path:
    return PACKAGE_PATH.joinpath(basename + suffix)


def run_pytest(
    pytester: pytest.Pytester, module: pathlib.Path, import_mode
) -> pytest.RunResult:
    return pytester.runpytest(
        "-vv",
        "--doctest-modules",
        "--doctest-cython",
        "--import-mode",
        import_mode,
        str(module),
    )


@pytest.mark.parametrize("import_mode", IMPORT_MODES)
def test_cython_ext_module(pytester, import_mode):
    module = get_module("cython_ext_module")
    assert module.exists()

    result = run_pytest(pytester, module, import_mode)
    result.stdout.fnmatch_lines(
        [
            "*Eggs.__init__*PASSED*",
            "*Eggs.blarg*PASSED*",
            "*Eggs.failing_test*FAILED*",
            "*Eggs.fubar*PASSED*",
            "*",
            "*FAILURES*",
            "*pypackage.cython_ext_module.Eggs.failing_test*",
            "078*",
            "079         >>> eggs = Eggs(1, 1)*",
            "080         >>> eggs.failing_test()*",
            "Expected:*",
            "    False*",
            "Got:*",
            "    True*",
        ]
    )
    assert result.ret == 1


@pytest.mark.parametrize("import_mode", IMPORT_MODES)
def test_wrap_c_ext_module(pytester, import_mode):
    module = get_module("wrap_c_ext_module")
    assert module.exists()

    result = run_pytest(pytester, module, import_mode)
    result.stdout.fnmatch_lines(
        [
            "*sqr*PASSED*",
        ]
    )
    assert result.ret == 0


@pytest.mark.parametrize("import_mode", IMPORT_MODES)
def test_wrap_cpp_ext_module(pytester, import_mode):
    module = get_module("wrap_cpp_ext_module")
    assert module.exists()

    result = run_pytest(pytester, module, import_mode)
    result.stdout.fnmatch_lines(
        [
            "*sqr*PASSED*",
        ]
    )
    assert result.ret == 0


@pytest.mark.parametrize("import_mode", IMPORT_MODES)
def test_pure_py_module(pytester, import_mode):
    module = get_module("pure_py_module", suffix=".py")
    assert module.exists()

    result = run_pytest(pytester, module, import_mode)
    result.stdout.fnmatch_lines(
        [
            "*Eggs.__init__*PASSED*",
            "*Eggs.foo*PASSED*",
            "*foo*PASSED*",
        ]
    )
    assert result.ret == 0




cdef int cfoo(int a, int b) except? -1
cdef int cbar(int a, int b) except? -1 nogil
cdef inline int cspam(int a, int b) except? -1 nogil


cdef class Eggs:
    cdef:
        readonly int a
        readonly int b

    cdef int foo(Eggs self) except? -1
    cdef int bar(Eggs self) except? -1 nogil
    cdef int spam(Eggs self) except? -1 nogil
    cpdef int fubar(Eggs self)



# distutils: language = c++

cdef extern from "../clib/sqrcpp.cpp":
    T vsqr[T](T a)


def sqr(a):
    """
    >>> sqr(2)
    4
    >>> '%.3g' % round(sqr(2.2), 2)
    '4.84'
    >>> sqr("asd")
    Traceback (most recent call last):
    ...
    TypeError: Expected int or float type input.
    """
    if isinstance(a, (int, )):
        return vsqr(<int>a)

    elif isinstance(a, (float, )):
        return vsqr(<double>a)

    else:
        raise TypeError("Expected int or float type input.")




cdef int cfoo(int a, int b) except? -1:
    """
    >>> cfoo(1, 1)
    2
    """
    return a + b

cdef int cbar(int a, int b) except? -1 nogil:
    """
    >>> cbar(1, 1)
    2
    """
    return a + b

cdef inline int cspam(int a, int b) except? -1 nogil:
    """
    >>> cspam(1, 1)
    2
    """
    return (a + b)


cdef class Eggs:

    def __init__(self, a, b):
        """
        >>> eggs = Eggs(1, 1)
        >>> eggs.a
        1
        >>> eggs.b
        1
        """
        self.a = a
        self.b = b

    cdef int foo(Eggs self) except? -1:
        """
        >>> eggs = Eggs(1, 1)
        >>> eggs.foo()
        2
        """
        return self.a + self.b

    cdef int bar(Eggs self) except? -1 nogil:
        """
        >>> eggs = Eggs(1, 1)
        >>> eggs.bar()
        2
        """
        return self.a + self.b

    cdef int spam(Eggs self) except? -1 nogil:
        """
        >>> eggs = Eggs(1, 1)
        >>> eggs.spam()
        2
        """
        return cspam(self.a, self.b)

    cpdef int fubar(Eggs self):
        """
        >>> eggs = Eggs(1, 1)
        >>> eggs.fubar()
        2
        """
        return self.a + self.b

    def blarg(self):
        """
        >>> eggs = Eggs(1, 1)
        >>> eggs.blarg()
        2
        """
        return self.a + self.b

    def failing_test(self):
        """
        >>> eggs = Eggs(1, 1)
        >>> eggs.failing_test()
        False
        """
        return True



#!/usr/bin/env python


if __name__ == "__main__":
    import sys

    from Cython.Build import cythonize
    from setuptools import Extension, setup

    directives = {
        "autotestdict": True,
        "embedsignature": False,
        "language_level": sys.version_info[0],
        "linetrace": False,
        "profile": False,
    }

    extensions = [Extension("*", ["src/pypackage/*.pyx"])]

    setup(
        name="pytest-cython-example",
        version="0.0.1",
        description="Example Cython project for pytest-cython tests",
        package_dir={"": "src"},
        packages=["pypackage"],
        zip_safe=False,
        ext_modules=cythonize(extensions, compiler_directives=directives),
    )
